"""Common utilities and boilerplate"""
import sys
import typing as t
from collections import namedtuple
from functools import wraps
from itertools import chain, starmap
from operator import attrgetter, methodcaller

__all__ = ["JSON", "Empty"]

T = t.TypeVar("T")
T1 = t.TypeVar("T1")
T2 = t.TypeVar("T2")


JSON = t.Union[
    str, int, float, bool, None, t.Dict[str, "JSON"], t.List["JSON"]
]


def identity(obj):
    return obj


class FrozenDict(t.Mapping[T1, T2]):
    # see https://stackoverflow.com/questions/45864273
    if not (3, 7) > sys.version_info > (3, 4):  # pragma: no cover
        __slots__ = "_inner"

    def __init__(self, inner):
        self._inner = inner if isinstance(inner, dict) else dict(inner)

    __len__ = property(attrgetter("_inner.__len__"))
    __iter__ = property(attrgetter("_inner.__iter__"))
    __getitem__ = property(attrgetter("_inner.__getitem__"))
    __repr__ = property(attrgetter("_inner.__repr__"))

    def __hash__(self):
        return hash(frozenset(self._inner.items()))


FrozenDict.EMPTY = FrozenDict({})


def merge(*dicts):
    """merge several mappings"""
    if dicts:
        return type(dicts[0])(
            chain.from_iterable(map(methodcaller("items"), dicts))
        )
    else:
        return {}


class Empty(Exception):
    """indicates a given list is unexpectedly empty"""


def init_last(items):
    # type: (t.List[T]) -> t.Tuple[t.List[T], T]
    """Return the first items and last item from a list

    Raises
    ------
    Empty
        if the given list is empty
    """
    try:
        return items[:-1], items[-1]
    except IndexError:
        raise Empty


def _make_init_fn(ntuple):
    @wraps(ntuple.__new__)
    def __init__(self, *args, **kwargs):
        self._values = ntuple(*args, **kwargs)

    return __init__


class _ValueObjectMeta(type(t.Generic)):
    """Metaclass for ``ValueObject``"""

    # TODO: add parameters to __doc__
    def __new__(self, name, bases, dct):
        # skip the ``ValueObject`` class itself
        if bases != (object,):
            fields = dct["__fields__"]
            fieldnames = [n for n, _, _ in dct["__fields__"]]
            assert "replace" not in fieldnames
            ntuple = namedtuple("_" + name, fieldnames)
            ntuple.__new__.__defaults__ = dct.get("__defaults__", ())
            dct.update(
                {
                    "__namedtuple_cls__": ntuple,
                    "__slots__": "_values",
                    # For the signature to appear correctly in
                    # introspection and docs,
                    # we create the __init__ function for
                    # each ValueObject class individually
                    "__init__": _make_init_fn(ntuple),
                }
            )
            dct.update(
                (name, property(attrgetter("_values." + name), doc=doc))
                for name, _, doc in fields
            )
        return super(_ValueObjectMeta, self).__new__(self, name, bases, dct)


class ValueObject(object, metaclass=_ValueObjectMeta):
    """Base class for "value object"-like classes,
    similar to frozen dataclasses in python 3.7+.

    Example
    -------

    >>> class Foo(ValueObject, ...):
    ...     __slots__ = '_values'  # optional
    ...     __fields__ = [
    ...         ('foo', int, 'the foo'),
    ...         ('bla', str, 'description for bla'),
    ...     ]
    ...
    >>> f = Foo(4, bla='foo')
    >>> f
    Foo(foo=4, bla='foo')

    """

    __slots__ = ()

    def replace(self, **kwargs):
        """Create a new instance, with certain fields replaced with new values

        Parameters
        ----------
        **kwargs
            Updated field values

        Example
        -------
        >>> my_object
        MyObject(a=5, b="qux")
        >>> my_object.replace(b="new!")
        MyObject(a=5, b="new!")

        """
        new = type(self).__new__(type(self))
        new._values = self._values._replace(**kwargs)
        return new

    def __eq__(self, other):
        if type(self) is type(other):
            return self._values == other._values
        return NotImplemented

    def __ne__(self, other):
        if type(self) is type(other):
            return self._values != other._values
        return NotImplemented

    def __repr__(self):
        try:
            return "{}({})".format(
                self.__class__.__qualname__,
                ", ".join(
                    starmap(
                        "{}={!r}".format,
                        zip(self._values._fields, self._values),
                    )
                ),
            )
        except Exception:
            return object.__repr__(self)

    __hash__ = property(attrgetter("_values.__hash__"))


class compose(object):
    """compose a function from a chain of functions
    Parameters
    ----------
    *funcs
        callables to compose
    Note
    ----
    * if given no functions, acts as an identity function
    """

    def __init__(self, *funcs):
        self.funcs = funcs
        self.__wrapped__ = funcs[-1] if funcs else identity

    def __call__(self, *args, **kwargs):
        if not self.funcs:
            return identity(*args, **kwargs)
        tail, head = self.funcs[:-1], self.funcs[-1]
        value = head(*args, **kwargs)
        for func in reversed(tail):
            value = func(value)
        return value



__version__ = "0.2.1"
__author__ = "Arie Bovenberg"
__copyright__ = "2019, Arie Bovenberg"
__license__ = "Apache 2.0"
__description__ = "Capable GraphQL client"

__all__ = [
    "__author__",
    "__version__",
    "__copyright__",
    "__license__",
    "__description__",
]



"""Root of the quiz API.

The entire public API is available at root level:

   >>> import quiz
   >>> quiz.Schema, quiz.execute, quiz.SelectionError, ...
"""
from .__about__ import *  # noqa
from .build import *  # noqa
from .execution import *  # noqa
from .schema import *  # noqa
from .types import *  # noqa

from . import build, schema  # isort:skip

__all__ = ["build", "schema"]



"""Functionality relating to the raw GraphQL schema"""
import enum
import json
import sys
import typing as t
from collections import defaultdict
from functools import partial
from itertools import chain
from operator import methodcaller
from types import new_class

from . import types
from .build import Query
from .compat import fspath
from .execution import execute
from .types import validate
from .utils import JSON, FrozenDict, ValueObject, merge

__all__ = ["Schema", "INTROSPECTION_QUERY"]

RawSchema = t.Dict[str, JSON]
ClassDict = t.Dict[str, type]


def _namedict(classes):
    return {c.__name__: c for c in classes}


def object_as_type(typ, interfaces, module):
    # type: (Object, t.Mapping[str, types.Interface], str) -> type
    # we don't add the fields yet -- these types may not exist yet.
    return type(
        str(typ.name),
        tuple(interfaces[i.name] for i in typ.interfaces) + (types.Object,),
        {"__doc__": typ.desc, "__raw__": typ, "__module__": module},
    )


def interface_as_type(typ, module):
    # type: (Interface, str) -> type
    # we don't add the fields yet -- these types may not exist yet.
    return new_class(
        str(typ.name),
        (types.Namespace,),
        kwds={"metaclass": types.Interface},
        exec_body=methodcaller(
            "update",
            {"__doc__": typ.desc, "__raw__": typ, "__module__": module},
        ),
    )


def enum_as_type(typ, module):
    # type: (Enum, str) -> Type[types.Enum]
    assert len(typ.values) > 0
    cls = types.Enum(
        typ.name, [(v.name, v.name) for v in typ.values], module=module
    )
    cls.__doc__ = typ.desc
    for member, conf in zip(cls.__members__.values(), typ.values):
        member.__doc__ = conf.desc
    return cls


def union_as_type(typ, objs):
    # type (Union, ClassDict) -> type
    assert len(typ.types) >= 1, "Encountered a Union with zero types"
    return type(
        str(typ.name),
        (types.Union,),
        {
            "__doc__": typ.desc,
            "__args__": tuple(objs[o.name] for o in typ.types),
        },
    )


def inputobject_as_type(typ):
    # type InputObject -> type
    return type(str(typ.name), (types.InputObject,), {"__doc__": typ.desc})


def _add_fields(obj, classes):
    for f in obj.__raw__.fields:
        setattr(
            obj,
            f.name,
            types.FieldDefinition(
                name=f.name,
                desc=f.desc,
                args=FrozenDict(
                    {
                        i.name: types.InputValue(
                            name=i.name,
                            desc=i.desc,
                            type=resolve_typeref(i.type, classes),
                        )
                        for i in f.args
                    }
                ),
                is_deprecated=f.is_deprecated,
                deprecation_reason=f.deprecation_reason,
                type=resolve_typeref(f.type, classes),
            ),
        )
    del obj.__raw__
    return obj


def resolve_typeref(ref, classes):
    # type: (TypeRef, ClassDict) -> type
    if ref.kind is Kind.NON_NULL:
        return _resolve_typeref_required(ref.of_type, classes)
    else:
        return types.Nullable[_resolve_typeref_required(ref, classes)]


def _resolve_typeref_required(ref, classes):
    assert ref.kind is not Kind.NON_NULL
    if ref.kind is Kind.LIST:
        return types.List[resolve_typeref(ref.of_type, classes)]
    return classes[ref.name]


class _QueryCreator(object):
    def __init__(self, schema):
        self.schema = schema

    def __getitem__(self, selection_set):
        cls = self.schema.query_type
        return Query(cls, selections=validate(cls, selection_set))


class Schema(ValueObject):
    """A GraphQL schema.

    Use :meth:`~Schema.from_path`, :meth:`~Schema.from_url`,
    or :meth:`~Schema.from_raw` to instantiate.
    """

    __fields__ = [
        ("classes", ClassDict, "Mapping of classes in the schema"),
        ("query_type", type, "The query type of the schema"),
        ("mutation_type", t.Optional[type], "The mutation type of the schema"),
        (
            "subscription_type",
            t.Optional[type],
            "The subscription type of the schema",
        ),
        (
            "module",
            t.Optional[str],
            "The module to which the classes are namespaced",
        ),
        ("raw", RawSchema, "The raw schema (JSON). To be deprecated"),
    ]

    def __getattr__(self, name):
        try:
            return self.classes[name]
        except KeyError:
            raise AttributeError(name)

    def __dir__(self):
        return list(self.classes) + dir(super(Schema, self))

    def populate_module(self):
        """Populate the schema's module with the schema's classes

        Note
        ----
        The schema's ``module`` must be set (i.e. not ``None``)

        Raises
        ------
        RuntimeError
            If the schema module is not set
        """
        # TODO: this is a bit ugly
        if self.module is None:
            raise RuntimeError("schema.module is not set")
        module_obj = sys.modules[self.module]
        for name, cls in self.classes.items():
            setattr(module_obj, name, cls)

    # interim object to allow slice syntax: Schema.query[...]
    @property
    def query(self):
        """Creator for a query operation

        Example
        -------

        >>> from quiz import SELECTOR as _
        >>> str(schema.query[
        ...     _
        ...     .field1
        ...     .foo
        ... ])
        query {
          field1
          foo
        }
        """
        return _QueryCreator(self)

    @classmethod
    def from_path(cls, path, module=None, scalars=()):
        """Create a :class:`Schema` from a JSON at a path

        Parameters
        ----------
        path: str or ~os.PathLike
            The path to the raw schema JSON file.
        module: ~typing.Optional[str], optional
            The name of the module to use when creating the schema's classes.
        scalars: ~typing.Iterable[~typing.Type[Scalar]]
            :class:`~quiz.types.Scalar` classes to use in the schema.
            Scalars in the schema, but not in this sequence, will be defined as
            :class:`~quiz.types.GenericScalar` subclasses.

        Returns
        -------
        Schema
            The generated schema

        Raises
        ------
        IOError
            If the file at given path cannot be read
        """
        with open(fspath(path)) as rfile:
            return cls.from_raw(
                json.load(rfile), module=module, scalars=scalars
            )

    def to_path(self, path):
        """Dump the schema as JSON to a path

        Parameters
        ----------
        path: str or ~os.PathLike
            The path to write the raw schema to
        """
        with open(fspath(path), "w") as wfile:
            json.dump(self.raw, wfile)

    @classmethod
    def from_raw(cls, raw_schema, module=None, scalars=()):
        """Create a :class:`Schema` from a raw JSON schema

        Parameters
        ----------
        raw_schema: ~typing.List[~typing.Dict[str, JSON]]
            The raw GraphQL schema.
            I.e. the result of the :data:`INTROSPECTION_QUERY`
        module: ~typing.Optional[str], optional
            The name of the module to use when creating classes
        scalars: ~typing.Iterable[~typing.Type[Scalar]]
            :class:`~quiz.types.Scalar` classes to use in the schema.
            Scalars in the schema, but not in this sequence, will be defined as
            :class:`~quiz.types.GenericScalar` subclasses.

        Returns
        -------
        Schema
            The schema constructed from raw data
        """
        by_kind = defaultdict(list)
        for tp in _load_types(raw_schema):
            by_kind[tp.__class__].append(tp)

        scalars_by_name = _namedict(scalars)
        scalars_by_name.update(types.BUILTIN_SCALARS)
        scalars_by_name.update(
            (
                tp.name,
                type(
                    str(tp.name), (types.GenericScalar,), {"__doc__": tp.desc}
                ),
            )
            for tp in by_kind[Scalar]
            if tp.name not in scalars_by_name
        )

        interfaces = _namedict(
            map(partial(interface_as_type, module=module), by_kind[Interface])
        )
        enums = _namedict(
            map(partial(enum_as_type, module=module), by_kind[Enum])
        )
        objs = _namedict(
            map(
                partial(object_as_type, interfaces=interfaces, module=module),
                by_kind[Object],
            )
        )
        unions = _namedict(
            map(partial(union_as_type, objs=objs), by_kind[Union])
        )
        input_objects = _namedict(
            map(inputobject_as_type, by_kind[InputObject])
        )

        classes = merge(
            scalars_by_name, interfaces, enums, objs, unions, input_objects
        )

        # we can only add fields after all classes have been created.
        for obj in chain(objs.values(), interfaces.values()):
            _add_fields(obj, classes)

        return cls(
            classes,
            query_type=classes[raw_schema["queryType"]["name"]],
            mutation_type=(
                raw_schema["mutationType"]
                and classes[raw_schema["mutationType"]["name"]]
            ),
            subscription_type=(
                raw_schema["subscriptionType"]
                and classes[raw_schema["subscriptionType"]["name"]]
            ),
            module=module,
            raw=raw_schema,
        )

    @classmethod
    def from_url(cls, url, scalars=(), module=None, **kwargs):
        """Build a GraphQL schema by introspecting an API

        Parameters
        ----------
        url: str
            URL of the target GraphQL API
        scalars: ~typing.Iterable[~typing.Type[Scalar]]
            :class:`~quiz.types.Scalar` classes to use in the schema.
            Scalars in the schema, but not in this sequence, will be defined as
            :class:`~quiz.types.GenericScalar` subclasses.

        module: ~typing.Optional[str], optional
            The module name to set on the generated classes
        **kwargs
            ``auth`` or ``client``, passed to :func:`~quiz.execution.execute`.

        Returns
        -------
        Schema
            The generated schema

        Raises
        ------
        ~quiz.types.ErrorResponse
            If there are errors in the response data
        """
        result = execute(INTROSPECTION_QUERY, url=url, **kwargs)
        return cls.from_raw(result["__schema"], scalars=scalars, module=module)

    # TODO: from_url_async


def _load_types(raw_schema):
    # type RawSchema -> Iterable[TypeSchema]
    return map(
        lambda typ: KIND_CAST[typ.kind](typ),
        map(_deserialize_type, raw_schema["types"]),
    )


INTROSPECTION_QUERY = """
{
  __schema {
    queryType { name }
    mutationType { name }
    subscriptionType { name }
    types {
      ...FullType
    }
  }
}

fragment FullType on __Type {
  kind
  name
  description
  fields(includeDeprecated: true) {
    name
    description
    args {
      ...InputValue
    }
    type {
      ...TypeRef
    }
    isDeprecated
    deprecationReason
  }
  inputFields {
    ...InputValue
  }
  interfaces {
    ...TypeRef
  }
  enumValues(includeDeprecated: true) {
    name
    description
    isDeprecated
    deprecationReason
  }
  possibleTypes {
    ...TypeRef
  }
}

fragment InputValue on __InputValue {
  name
  description
  type { ...TypeRef }
  defaultValue
}

fragment TypeRef on __Type {
  kind
  name
  ofType {
    kind
    name
    ofType {
      kind
      name
      ofType {
        kind
        name
      }
    }
  }
}
"""
"""Query to retrieve the raw schema"""


class Kind(enum.Enum):
    OBJECT = "OBJECT"
    SCALAR = "SCALAR"
    NON_NULL = "NON_NULL"
    LIST = "LIST"
    INTERFACE = "INTERFACE"
    ENUM = "ENUM"
    INPUT_OBJECT = "INPUT_OBJECT"
    UNION = "UNION"


KIND_CAST = {
    Kind.SCALAR: lambda typ: Scalar(name=typ.name, desc=typ.desc),
    Kind.OBJECT: lambda typ: Object(
        name=typ.name,
        desc=typ.desc,
        interfaces=typ.interfaces,
        input_fields=typ.input_fields,
        fields=typ.fields,
    ),
    Kind.INTERFACE: lambda typ: Interface(
        name=typ.name, desc=typ.desc, fields=typ.fields
    ),
    Kind.ENUM: lambda typ: Enum(
        name=typ.name, desc=typ.desc, values=typ.enum_values
    ),
    Kind.UNION: lambda typ: Union(
        name=typ.name, desc=typ.desc, types=typ.possible_types
    ),
    Kind.INPUT_OBJECT: lambda typ: InputObject(
        name=typ.name, desc=typ.desc, input_fields=typ.input_fields
    ),
}


TypeRef = t.NamedTuple(
    "TypeRef",
    [
        ("name", t.Optional[str]),
        ("kind", Kind),
        ("of_type", t.Optional["TypeRef"]),
    ],
)
InputValue = t.NamedTuple(
    "InputValue",
    [("name", str), ("desc", str), ("type", TypeRef), ("default", object)],
)
Field = t.NamedTuple(
    "Field",
    [
        ("name", str),
        ("type", TypeRef),
        ("args", t.List[InputValue]),
        ("desc", str),
        ("is_deprecated", bool),
        ("deprecation_reason", t.Optional[str]),
    ],
)
Type = t.NamedTuple(
    "Type",
    [
        ("name", t.Optional[str]),
        ("kind", Kind),
        ("desc", str),
        ("fields", t.Optional[t.List[Field]]),
        ("input_fields", t.Optional[t.List["InputValue"]]),
        ("interfaces", t.Optional[t.List[TypeRef]]),
        ("possible_types", t.Optional[t.List[TypeRef]]),
        ("enum_values", t.Optional[t.List]),
    ],
)
EnumValue = t.NamedTuple(
    "EnumValue",
    [
        ("name", str),
        ("desc", str),
        ("is_deprecated", bool),
        ("deprecation_reason", t.Optional[str]),
    ],
)


def make_inputvalue(conf):
    return InputValue(
        name=conf["name"],
        desc=conf["description"],
        type=make_typeref(conf["type"]),
        default=conf["defaultValue"],
    )


def make_typeref(conf):
    return TypeRef(
        name=conf["name"],
        kind=Kind(conf["kind"]),
        of_type=conf.get("ofType") and make_typeref(conf["ofType"]),
    )


def make_field(conf):
    return Field(
        name=conf["name"],
        type=make_typeref(conf["type"]),
        args=list(map(make_inputvalue, conf["args"])),
        desc=conf["description"],
        is_deprecated=conf["isDeprecated"],
        deprecation_reason=conf["deprecationReason"],
    )


def make_enumval(conf):
    return EnumValue(
        name=conf["name"],
        desc=conf["description"],
        is_deprecated=conf["isDeprecated"],
        deprecation_reason=conf["deprecationReason"],
    )


def _deserialize_type(conf):
    # type: (t.Dict[str, JSON]) -> Type
    return Type(
        name=conf["name"],
        kind=Kind(conf["kind"]),
        desc=conf["description"],
        fields=conf["fields"] and list(map(make_field, conf["fields"])),
        input_fields=conf["inputFields"]
        and list(map(make_inputvalue, conf["inputFields"])),
        interfaces=conf["interfaces"]
        and list(map(make_typeref, conf["interfaces"])),
        possible_types=conf["possibleTypes"]
        and list(map(make_typeref, conf["possibleTypes"])),
        enum_values=conf["enumValues"]
        and list(map(make_enumval, conf["enumValues"])),
    )


Interface = t.NamedTuple(
    "Interface", [("name", str), ("desc", str), ("fields", t.List[Field])]
)
Object = t.NamedTuple(
    "Object",
    [
        ("name", str),
        ("desc", str),
        ("interfaces", t.List[TypeRef]),
        ("input_fields", t.Optional[t.List[InputValue]]),
        ("fields", t.List[Field]),
    ],
)
Scalar = t.NamedTuple("Scalar", [("name", str), ("desc", str)])
Enum = t.NamedTuple(
    "Enum", [("name", str), ("desc", str), ("values", t.List[EnumValue])]
)
Union = t.NamedTuple(
    "Union", [("name", str), ("desc", str), ("types", t.List[TypeRef])]
)
InputObject = t.NamedTuple(
    "InputObject",
    [("name", str), ("desc", str), ("input_fields", t.List[InputValue])],
)
TypeSchema = t.Union[Interface, Object, Scalar, Enum, Union, InputObject]



import os
import sys
import typing as t

HAS_PEP519 = sys.version_info > (3, 6)


if HAS_PEP519:
    from os import fspath
else:  # pragma: no cover
    # code adapted from https://www.python.org/dev/peps/pep-0519/
    def fspath(path):
        # type: (t.Union[os.Pathlike, str, bytes]) -> t.Union[str, bytes]
        """Return the string representation of the path.

        If str or bytes is passed in, it is returned unchanged. If __fspath__()
        returns something other than str or bytes then TypeError is raised. If
        this function is given something that is not str, bytes, or os.PathLike
        then TypeError is raised.
        """
        if isinstance(path, (str, bytes)):
            return path

        # Work from the object's type to match method resolution of other magic
        # methods.
        path_type = type(path)
        try:
            path = path_type.__fspath__(path)
        except AttributeError:
            if hasattr(path_type, "__fspath__"):
                raise
        else:
            if isinstance(path, (str, bytes)):
                return path
            else:
                raise TypeError(
                    "expected __fspath__() to return str or bytes, "
                    "not " + type(path).__name__
                )

        raise TypeError(
            "expected str, bytes or os.PathLike object, not "
            + path_type.__name__
        )



"""Components for executing GraphQL operations"""
import json
import typing as t
from functools import partial

import snug
from gentools import irelay

from .build import Query
from .types import load
from .utils import JSON, ValueObject

__all__ = [
    "execute",
    "execute_async",
    "executor",
    "async_executor",
    "Executable",
    "ErrorResponse",
    "HTTPError",
    "RawResult",
    "QueryMetadata",
]

Executable = t.Union[str, Query]
"""Anything which can be executed as a GraphQL operation"""


def _exec(executable):
    # type: (Executable) -> t.Generator
    if isinstance(executable, str):
        return (yield executable)
    elif isinstance(executable, Query):
        return load(
            executable.cls, executable.selections, (yield str(executable))
        )
    else:
        raise NotImplementedError("not executable: " + repr(executable))


def middleware(url, query_str):
    # type: (str, str) -> snug.Query[t.Dict[str, JSON]]
    request = snug.POST(
        url,
        content=json.dumps({"query": query_str}).encode("ascii"),
        headers={"Content-Type": "application/json"},
    )
    response = yield request
    if response.status_code >= 400:
        raise HTTPError(response, request)
    content = json.loads(response.content.decode("utf-8"))
    if "errors" in content:
        content.setdefault("data", {})
        raise ErrorResponse(**content)
    return RawResult(
        content["data"], QueryMetadata(request=request, response=response)
    )


def execute(obj, url, **kwargs):
    """Execute a GraphQL executable

    Parameters
    ----------
    obj: :data:`~quiz.execution.Executable`
        The object to execute.
        This may be a raw string or a query
    url: str
        The URL of the target endpoint
    **kwargs
         ``auth`` and/or ``client``, passed to :func:`snug.query.execute`.

    Returns
    -------
    RawResult (a dict) or the schema's return type
        In case of a raw string, a raw result.
        Otherwise, an instance of the schema's type queried for.

    Raises
    ------
    ErrorResponse
        If errors are present in the response
    HTTPError
        If the response has a non 2xx response code
    """
    snug_query = irelay(_exec(obj), partial(middleware, url))
    return snug.execute(snug_query, **kwargs)


def executor(**kwargs):
    """Create a version of :func:`execute` with bound arguments.
    Equivalent to ``partial(execute, **kwargs)``.

    Parameters
    ----------
    **kwargs
       ``url``, ``auth``, and/or ``client``, passed to :func:`execute`

    Returns
    -------
    ~typing.Callable[[Executable], JSON]
        A callable to execute GraphQL executables

    Example
    -------

    >>> execute = executor(url='https://api.github.com/graphql',
    ...                    auth=('me', 'password'))
    >>> result = execute('''
    ...   {
    ...     repository(owner: "octocat" name: "Hello-World") {
    ...       description
    ...     }
    ...   }
    ... ''', client=requests.Session())
    """
    return partial(execute, **kwargs)


def execute_async(obj, url, **kwargs):
    """Execute a GraphQL executable asynchronously

    Parameters
    ----------
    obj: Executable
        The object to execute.
        This may be a raw string or a query
    url: str
        The URL of the target endpoint
    **kwargs
         ``auth`` and/or ``client``,
         passed to :func:`snug.query.execute_async`.

    Returns
    -------
    RawResult (a dict) or the schema's return type
        In case of a raw string, a raw result.
        Otherwise, an instance of the schema's type queried for.


    Raises
    ------
    ErrorResponse
        If errors are present in the response
    HTTPError
        If the response has a non 2xx response code
    """
    snug_query = irelay(_exec(obj), partial(middleware, url))
    return snug.execute_async(snug_query, **kwargs)


def async_executor(**kwargs):
    """Create a version of :func:`execute_async` with bound arguments.
    Equivalent to ``partial(execute_async, **kwargs)``.

    Parameters
    ----------
    **kwargs
       ``url``, ``auth``, and/or ``client``, passed to :func:`execute_async`

    Returns
    -------
    ~typing.Callable[[Executable], ~typing.Awaitable[JSON]]
        A callable to asynchronously execute GraphQL executables

    Example
    -------

    >>> execute = async_executor(url='https://api.github.com/graphql',
    ...                          auth=('me', 'password'))
    >>> result = await execute('''
    ...   {
    ...     repository(owner: "octocat" name: "Hello-World") {
    ...       description
    ...     }
    ...   }
    ... ''')
    """
    return partial(execute_async, **kwargs)


class ErrorResponse(ValueObject, Exception):
    """A response containing errors"""

    __fields__ = [
        ("data", t.Dict[str, JSON], "Data returned in the response"),
        (
            "errors",
            t.List[t.Dict[str, JSON]],
            "Errors returned in the response",
        ),
    ]


class RawResult(dict):
    """Dictionary as result of a raw query.

    Contains HTTP :class:`metadata <QueryMetadata>`
    in its ``__metadata__`` attribute.
    """

    __slots__ = "__metadata__"

    def __init__(self, items, meta):
        super(RawResult, self).__init__(items)
        self.__metadata__ = meta


class QueryMetadata(ValueObject):
    """HTTP metadata for query"""

    __fields__ = [
        ("response", snug.Response, "The response object"),
        ("request", snug.Request, "The original request"),
    ]


class HTTPError(ValueObject, Exception):
    """Indicates a response with a non 2xx status code"""

    __fields__ = [
        ("response", snug.Response, "The response object"),
        ("request", snug.Request, "The original request"),
    ]

    def __str__(self):
        return (
            "Response with status {0.status_code}, content: {0.content!r} "
            'for URL "{1.url}". View this exception\'s `request` and '
            "`response` attributes for detailed info.".format(
                self.response, self.request
            )
        )



import os.path

from setuptools import find_packages, setup


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    with open(path, "r") as rfile:
        return rfile.read()


metadata = {}
exec(read("quiz/__about__.py"), metadata)


setup(
    name="quiz",
    version=metadata["__version__"],
    description=metadata["__description__"],
    license=metadata["__license__"],
    long_description=read("README.rst") + "\n\n" + read("HISTORY.rst"),
    long_description_content_type='text/plain',
    url="https://github.com/ariebovenberg/quiz",
    author=metadata["__author__"],
    author_email="a.c.bovenberg@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=["gentools~=1.1.0", "snug~=2.0.0"],
    keywords=["graphql", "http", "async"],
    python_requires=">=3.6",
    packages=find_packages(exclude=("tests", "docs")),
)



"""Components for typed GraphQL interactions"""
import enum
import typing as t
from itertools import starmap

from .build import Field, InlineFragment, SelectionSet
from .utils import JSON, FrozenDict, ValueObject

__all__ = [
    # types
    "Enum",
    "Union",
    "GenericScalar",
    "Scalar",
    "List",
    "Interface",
    "Object",
    "Nullable",
    "FieldDefinition",
    "InputValue",
    # TODO: mutation
    # TODO: subscription
    # validation
    "validate",
    "ValidationError",
    "SelectionError",
    "NoSuchField",
    "NoSuchArgument",
    "SelectionsNotSupported",
    "InvalidArgumentType",
    "MissingArgument",
    "NoValueForField",
    "load",
]


InputValue = t.NamedTuple(
    "InputValue", [("name", str), ("desc", str), ("type", type)]
)

_PRIMITIVE_TYPES = (int, float, bool, str)


class HasFields(type):
    """metaclass for classes with GraphQL field definitions"""

    def __getitem__(self, selection_set):
        # type: (SelectionSet) -> InlineFragment
        return InlineFragment(self, validate(self, selection_set))


class Namespace(object):
    def __init__(__self, **kwargs):
        __self.__dict__.update(kwargs)

    def __fields__(self):
        return {k: v for k, v in self.__dict__.items() if k != "__metadata__"}

    def __eq__(self, other):
        if type(self) == type(other):
            return self.__fields__() == other.__fields__()
        return NotImplemented

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__qualname__,
            ", ".join(starmap("{}={!r}".format, self.__dict__.items())),
        )


class Object(Namespace, metaclass=HasFields):
    """a graphQL object"""


class InputObject(object):
    """not yet implemented"""


# separate class to distinguish graphql enums from normal Enums
class Enum(enum.Enum):
    pass


class Interface(HasFields):
    """metaclass for interfaces"""


class NoValueForField(AttributeError):
    """Indicates a value cannot be retrieved for the field"""


class FieldDefinition(ValueObject):
    __fields__ = [
        ("name", str, "Field name"),
        ("desc", str, "Field description"),
        ("type", type, "Field data type"),
        ("args", FrozenDict[str, InputValue], "Accepted field arguments"),
        ("is_deprecated", bool, "Whether the field is deprecated"),
        ("deprecation_reason", t.Optional[str], "Reason for deprecation"),
    ]

    def __get__(self, obj, objtype=None):
        if obj is None:  # accessing on class
            return self
        try:
            return obj.__dict__[self.name]
        except KeyError:
            raise NoValueForField()

    # full descriptor interface is necessary to be displayed nicely in help()
    def __set__(self, obj, value):
        raise AttributeError("Can't set field value")

    # __doc__ allows descriptor to be displayed nicely in help()
    @property
    def __doc__(self):
        return ": {.__name__}\n    {}".format(self.type, self.desc)


class ListMeta(type):
    def __getitem__(self, arg):
        return type("[{.__name__}]".format(arg), (List,), {"__arg__": arg})

    def __instancecheck__(self, instance):
        return isinstance(instance, list) and all(
            isinstance(i, self.__arg__) for i in instance
        )


# Q: why not typing.List?
# A: it doesn't support __doc__, __name__, or isinstance()
class List(object, metaclass=ListMeta):
    __arg__ = object


class NullableMeta(type):
    def __getitem__(self, arg):
        return type(
            "{.__name__} or None".format(arg), (Nullable,), {"__arg__": arg}
        )

    def __instancecheck__(self, instance):
        return instance is None or isinstance(instance, self.__arg__)


# Q: why not typing.Optional?
# A: it is not easily distinguished from Union,
#    and doesn't support __doc__, __name__, or isinstance()
class Nullable(object, metaclass=NullableMeta):
    __arg__ = object


class UnionMeta(type):
    def __instancecheck__(self, instance):
        return isinstance(instance, self.__args__)


# Q: why not typing.Union?
# A: it isn't consistent across python versions,
#    and doesn't support __doc__, __name__, or isinstance()
class Union(object, metaclass=UnionMeta):
    __args__ = ()


class Scalar(object):
    """Base class for scalars"""

    def __gql_dump__(self):
        """Serialize the scalar to a GraphQL primitive value"""
        raise NotImplementedError(
            "GraphQL serialization is not defined for this scalar"
        )

    @classmethod
    def __gql_load__(cls, data):
        """Load a scalar instance from GraphQL"""
        raise NotImplementedError(
            "GraphQL deserialization is not defined for this scalar"
        )


class GenericScalarMeta(type):
    def __instancecheck__(self, instance):
        return isinstance(instance, _PRIMITIVE_TYPES)


class GenericScalar(Scalar, metaclass=GenericScalarMeta):
    """A generic scalar, accepting any primitive type"""


def _unwrap_list_or_nullable(type_):
    # type: t.Type[Nullable, List, Scalar, Enum, InputObject]
    # -> Type[Scalar | Enum | InputObject]
    if issubclass(type_, (Nullable, List)):
        return _unwrap_list_or_nullable(type_.__arg__)
    return type_


def _validate_args(schema, actual):
    # type: (t.Mapping[str, InputValue], t.Mapping[str, object])
    # -> Mapping[str, object]
    invalid_args = actual.keys() - schema.keys()
    if invalid_args:
        raise NoSuchArgument(invalid_args.pop())

    for input_value in schema.values():
        try:
            value = actual[input_value.name]
        except KeyError:
            if issubclass(input_value.type, Nullable):
                continue  # arguments of nullable type may be omitted
            else:
                raise MissingArgument(input_value.name)

        if not isinstance(value, input_value.type):
            raise InvalidArgumentType(input_value.name, value)

    return actual


def _validate_field(schema, actual):
    # type (Optional[FieldDefinition], Field) -> Field
    # raises:
    # - NoSuchField
    # - SelectionsNotSupported
    # - NoSuchArgument
    # - RequredArgument
    # - InvalidArgumentType
    if schema is None:
        raise NoSuchField()
    _validate_args(schema.args, actual.kwargs)
    if actual.selection_set:
        type_ = _unwrap_list_or_nullable(schema.type)
        if not isinstance(type_, HasFields):
            raise SelectionsNotSupported()
        validate(type_, actual.selection_set)
    return actual


def validate(cls, selection_set):
    """Validate a selection set against a type

    Parameters
    ----------
    cls: type
        The class to validate against, an ``Object`` or ``Interface``
    selection_set: SelectionSet
        The selection set to validate

    Returns
    -------
    SelectionSet
        The validated selection set

    Raises
    ------
    SelectionError
        If the selection set is not valid
    """
    for field in selection_set:
        try:
            _validate_field(getattr(cls, field.name, None), field)
        except ValidationError as e:
            raise SelectionError(cls, field.name, e)
    return selection_set


T = t.TypeVar("T")


# TODO: refactor using singledispatch
# TODO: cleanup this API: ``field`` is often unneeded. unify with ``load``?
def load_field(type_, field, value):
    # type: (t.Type[T], Field, JSON) -> T
    if issubclass(type_, Namespace):
        assert isinstance(value, dict)
        return load(type_, field.selection_set, value)
    elif issubclass(type_, Nullable):
        return (
            None if value is None else load_field(type_.__arg__, field, value)
        )
    elif issubclass(type_, List):
        assert isinstance(value, list)
        return [load_field(type_.__arg__, field, v) for v in value]
    elif issubclass(type_, _PRIMITIVE_TYPES):
        assert isinstance(value, type_)
        return value
    elif issubclass(type_, GenericScalar):
        assert isinstance(value, type_)
        return value
    elif issubclass(type_, Scalar):
        return type_.__gql_load__(value)
    elif issubclass(type_, Enum):
        assert value, type_._members_names_
        return type_(value)
    else:
        raise NotImplementedError()


def load(cls, selection_set, response):
    """Load a response for a selection set

    Parameters
    ----------
    cls: Type[T]
        The class to load against, an ``Object`` or ``Interface``
    selection_set: SelectionSet
        The selection set to validate
    response: t.Mapping[str, JSON]
        The JSON response data

    Returns
    -------
    T
        An instance of ``cls``
    """
    instance = cls(
        **{
            field.alias
            or field.name: load_field(
                getattr(cls, field.name).type,
                field,
                response[field.alias or field.name],
            )
            for field in selection_set
        }
    )
    # TODO: do this in a cleaner way
    if hasattr(response, "__metadata__"):
        instance.__metadata__ = response.__metadata__
    return instance


class ValidationError(Exception):
    """base class for validation errors"""


class SelectionError(ValueObject, ValidationError):
    __fields__ = [
        ("on", type, "Type on which the error occurred"),
        ("path", str, "Path at which the error occurred"),
        ("error", ValidationError, "Original error"),
    ]

    def __str__(self):
        return '{} on "{}" at path "{}":\n\n    {}: {}'.format(
            self.__class__.__name__,
            self.on.__name__,
            self.path,
            self.error.__class__.__name__,
            self.error,
        )


class NoSuchField(ValueObject, ValidationError):
    __fields__ = []

    def __str__(self):
        return "field does not exist"


class NoSuchArgument(ValueObject, ValidationError):
    __fields__ = [("name", str, "(Invalid) argument name")]

    def __str__(self):
        return 'argument "{}" does not exist'.format(self.name)


class InvalidArgumentType(ValueObject, ValidationError):
    __fields__ = [
        ("name", str, "Argument name"),
        ("value", object, "(Invalid) value"),
    ]

    def __str__(self):
        return 'invalid value "{}" of type {} for argument "{}"'.format(
            self.value, type(self.value), self.name
        )


class MissingArgument(ValueObject, ValidationError):
    __fields__ = [("name", str, "Missing argument name")]

    def __str__(self):
        return 'argument "{}" missing (required)'.format(self.name)


class SelectionsNotSupported(ValueObject, ValidationError):
    __fields__ = []

    def __str__(self):
        return "selections not supported on this object"


BUILTIN_SCALARS = {"Boolean": bool, "String": str, "Float": float, "Int": int}
